library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;  -- Use OR_REDUCE function

use work.v1495pkg.all;
use work.MTriggerPKG.all;

entity MTrigger is
	port( 
		nLBRES      : IN     std_logic;                       -- Async Reset (active low)
		LCLK        : IN     std_logic;                       -- Local Bus Clock
		--*************************************************
		-- REGISTER INTERFACE
		--*************************************************
		REG_WREN    : IN     std_logic;                       -- Write pulse (active high)
		REG_RDEN    : IN     std_logic;                       -- Read  pulse (active high)
		REG_ADDR    : IN     std_logic_vector (15 DOWNTO 0);  -- Register address
		REG_DIN     : IN     std_logic_vector (15 DOWNTO 0);  -- Data from CAEN Local Bus
		REG_DOUT    : OUT    std_logic_vector (15 DOWNTO 0);  -- Data to   CAEN Local Bus
		USR_ACCESS  : IN     std_logic;                       -- Current register access is 
													-- at user address space(Active high)
		--*************************************************
		-- V1495 Front Panel Ports (PORT A,B,C,G)
		--*************************************************
		A_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- In A (32 x LVDS/ECL)
		B_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- In B (32 x LVDS/ECL) 
		C_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- Out C (32 x LVDS)
		G_LEV       : OUT    std_logic;                       -- Output Level Select (NIM/TTL)
		G_DIR       : OUT    std_logic;                       -- Output Enable
		G_DOUT      : OUT    std_logic_vector (1 DOWNTO 0);   -- Out G - LEMO (2 x NIM/TTL)
		G_DIN       : IN     std_logic_vector (1 DOWNTO 0);   -- In G - LEMO (2 x NIM/TTL)
		--*************************************************
		-- A395x MEZZANINES INTERFACES (PORT D,E,F)
		--*************************************************
		-- Expansion Mezzanine Identifier:
		-- x_IDCODE :
		-- 000 : A395A (32 x IN LVDS/ECL)
		-- 001 : A395B (32 x OUT LVDS)
		-- 010 : A395C (32 x OUT ECL)
		-- 011 : A395D (8  x IN/OUT NIM/TTL)

		-- Expansion Mezzanine Port Signal Standard Select
		-- x_LEV : 
		--    0=>TTL,1=>NIM

		-- Expansion Mezzanine Port Direction
		-- x_DIR : 
		--    0=>OUT,1=>IN

		-- In/Out D (I/O Expansion)
		D_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- D slot mezzanine Identifier
		D_LEV       : OUT    std_logic;                       -- D slot Port Signal Level Select 
		D_DIR       : OUT    std_logic;                       -- D slot Port Direction
		D_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- D slot Data In  Bus
		D_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- D slot Data Out Bus
		-- In/Out E (I/O Expansion)
		E_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- E slot mezzanine Identifier
		E_LEV       : OUT    std_logic;                       -- E slot Port Signal Level Select
		E_DIR       : OUT    std_logic;                       -- E slot Port Direction
		E_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- E slot Data In  Bus
		E_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- E slot Data Out Bus
		-- In/Out F (I/O Expansion)
		F_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- F slot mezzanine Identifier
		F_LEV       : OUT    std_logic;                       -- F slot Port Signal Level Select
		F_DIR       : OUT    std_logic;                       -- F slot Port Direction
		F_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- F slot Data In  Bus
		F_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- F slot Data Out Bus
		--*************************************************
		-- DELAY LINES
		--*************************************************
		-- PDL = Programmable Delay Lines  (Step = 0.25ns / FSR = 64ns)
		-- DLO = Delay Line Oscillator     (Half Period ~ 10 ns)
		-- 3D3428 PDL (PROGRAMMABLE DELAY LINE) CONFIGURATION
		PDL_WR      : OUT    std_logic;                       -- Write Enable
		PDL_SEL     : OUT    std_logic;                       -- PDL Selection (0=>PDL0, 1=>PDL1)
		PDL_READ    : IN     std_logic_vector ( 7 DOWNTO 0);  -- Read Data
		PDL_WRITE   : OUT    std_logic_vector ( 7 DOWNTO 0);  -- Write Data
		PDL_DIR     : OUT    std_logic;                       -- Direction (0=>Write, 1=>Read)
		-- DELAY I/O
		PDL0_OUT    : IN     std_logic;                       -- Signal from PDL0 Output
		PDL1_OUT    : IN     std_logic;                       -- Signal from PDL1 Output
		DLO0_OUT    : IN     std_logic;                       -- Signal from DLO0 Output
		DLO1_OUT    : IN     std_logic;                       -- Signal from DLO1 Output
		PDL0_IN     : OUT    std_logic;                       -- Signal to   PDL0 Input
		PDL1_IN     : OUT    std_logic;                       -- Signal to   PDL1 Input
		DLO0_GATE   : OUT    std_logic;                       -- DLO0 Gate (active high)
		DLO1_GATE   : OUT    std_logic;                       -- DLO1 Gate (active high)
		--*************************************************
		-- SPARE PORTS
		--*************************************************
		SPARE_OUT    : OUT   std_logic_vector(11 downto 0);   -- SPARE Data Out 
		SPARE_IN     : IN    std_logic_vector(11 downto 0);   -- SPARE Data In
		SPARE_DIR    : OUT   std_logic_vector(11 downto 0);   -- SPARE Direction (0 => OUT, 1 => IN)   
		--*************************************************
		-- LED
		--*************************************************
		RED_PULSE       : OUT    std_logic;                   -- RED   Led Pulse (active high)
		GREEN_PULSE     : OUT    std_logic                    -- GREEN Led Pulse (active high)
	);
end MTrigger;

architecture rtl of MTrigger is

	component MPLLBlock
		port(
			areset		: IN STD_LOGIC  := '0';
			inclk0		: IN STD_LOGIC  := '0';
			c0			: OUT STD_LOGIC ;
			locked		: OUT STD_LOGIC 
		);
	end component;

	component QISectSum
		port(
			CLK		: in std_logic;
			QA		: in std_logic_vector(54 downto 0);
			QB		: in std_logic_vector(54 downto 0);
			QC		: in std_logic_vector(54 downto 0);
			QSUMB	: out STD_LOGIC_VEC_ARRAY4_37
		);
	end component;
	
	component QISectClusters
		port(
			CLK			: in std_logic;
			
			QSUMA		: in STD_LOGIC_VEC_ARRAY4_37;
			QSUMB		: in STD_LOGIC_VEC_ARRAY4_37;
			QSUMC		: in STD_LOGIC_VEC_ARRAY4_37;
			
			THRESHOLD	: in std_logic_vector(3 downto 0);
			
			QICLUSTERSB	: in std_logic_vector(18 downto 0);
			QOCLUSTERSB	: out std_logic_vector(36 downto 0)
		);
	end component;

	component ParallelComRx
		port(
			CLK			: in std_logic;
			RESET		: in std_logic;
			
			DIN			: in std_logic_vector(31 downto 0);
						
			HITBITS		: out std_logic_vector(54 downto 0);
			CLUSTERS	: out std_logic_vector(18 downto 0);
			CLUSTERCNT1	: out std_logic_vector(3 downto 0);
			CLUSTERCNT2	: out std_logic_vector(3 downto 0);
			
			COMERROR	: out std_logic;
			
			SYNC		: in std_logic
		);
	end component;
	
	component TriggerDetect
		port(
				CLK				: in std_logic;
				RESET			: in std_logic;

				SYNC			: in std_logic;
				
				Q1_CLUSTERCNT1	: in std_logic_vector(3 downto 0);
				Q1_CLUSTERCNT2	: in std_logic_vector(3 downto 0);
				Q1_CLUSTERS		: in std_logic_vector(36 downto 0);

				Q2_CLUSTERCNT1	: in std_logic_vector(3 downto 0);
				Q2_CLUSTERCNT2	: in std_logic_vector(3 downto 0);
				Q2_CLUSTERS		: in std_logic_vector(36 downto 0);

				Q3_CLUSTERCNT1	: in std_logic_vector(3 downto 0);
				Q3_CLUSTERCNT2	: in std_logic_vector(3 downto 0);
				Q3_CLUSTERS		: in std_logic_vector(36 downto 0);

				Q4_CLUSTERCNT1	: in std_logic_vector(3 downto 0);
				Q4_CLUSTERCNT2	: in std_logic_vector(3 downto 0);
				Q4_CLUSTERS		: in std_logic_vector(36 downto 0);
				
				CLUSTERCNT_THR	: in std_logic_vector(6 downto 0);
				TRIGGER			: out std_logic
		);
	end component;
	
	signal PLL_LOCK					: std_logic;
	signal SYS_RESET				: std_logic;	
	signal RESET					: std_logic;
	signal ERROR					: std_logic;
	signal CARD_D_ERROR				: std_logic;
	signal CARD_E_ERROR				: std_logic;
	signal CARD_F_ERROR				: std_logic;
	
	signal HEART_BEAT_CNT_D			: std_logic_vector(25 downto 0);
	signal HEART_BEAT_CNT_Q			: std_logic_vector(25 downto 0);
	
	signal PLLCLK					: std_logic;
	
	signal REFCLK_D, REFCLK_Q		: std_logic;
	
	signal Q1_HITBITS, Q2_HITBITS	: std_logic_vector(54 downto 0);
	signal Q3_HITBITS, Q4_HITBITS	: std_logic_vector(54 downto 0);
	
	signal Q1_ICLUSTERS				: std_logic_vector(18 downto 0);
	signal Q2_ICLUSTERS				: std_logic_vector(18 downto 0);
	signal Q3_ICLUSTERS				: std_logic_vector(18 downto 0);
	signal Q4_ICLUSTERS				: std_logic_vector(18 downto 0);
	
	signal Q1_CLUSTERS, Q2_CLUSTERS	: std_logic_vector(36 downto 0);
	signal Q3_CLUSTERS, Q4_CLUSTERS	: std_logic_vector(36 downto 0);
	
	signal Q1_CLUSTERCNT1			: std_logic_vector(3 downto 0);
	signal Q2_CLUSTERCNT1			: std_logic_vector(3 downto 0);
	signal Q3_CLUSTERCNT1			: std_logic_vector(3 downto 0);
	signal Q4_CLUSTERCNT1			: std_logic_vector(3 downto 0);

	signal Q1_CLUSTERCNT2			: std_logic_vector(3 downto 0);
	signal Q2_CLUSTERCNT2			: std_logic_vector(3 downto 0);
	signal Q3_CLUSTERCNT2			: std_logic_vector(3 downto 0);
	signal Q4_CLUSTERCNT2			: std_logic_vector(3 downto 0);
	
	signal Q1_SUM, Q2_SUM			: STD_LOGIC_VEC_ARRAY4_37;
	signal Q3_SUM, Q4_SUM			: STD_LOGIC_VEC_ARRAY4_37;

	signal CLUSTER_THRESHOLD		: std_logic_vector(3 downto 0);
	signal CLUSTERCNT_THR			: std_logic_vector(6 downto 0);
	
	signal SYNC_D, SYNC_Q			: std_logic_vector(2 downto 0);
	signal SYNC						: std_logic;
	
	signal Q1_COMERROR, Q2_COMERROR	: std_logic;
	signal Q3_COMERROR, Q4_COMERROR	: std_logic;
	
	signal TRIGGER					: std_logic;
BEGIN

	RED_PULSE <= ERROR;
	GREEN_PULSE <= HEART_BEAT_CNT_Q(25);

	SYS_RESET <= not nLBRES;
	RESET <= not PLL_LOCK;
--	ERROR <= RESET or CARD_D_ERROR or CARD_E_ERROR or CARD_F_ERROR or Q1_COMERROR or Q2_COMERROR or Q3_COMERROR or Q4_COMERROR;
	ERROR <= RESET or Q1_COMERROR;

	CARD_D_ERROR <= '1' when D_IDCODE /= A395A else '0';
	CARD_E_ERROR <= '1' when E_IDCODE /= A395A else '0';
	CARD_F_ERROR <= '1' when E_IDCODE /= A395D else '0';

	D_DIR <= '1';	-- Input
	E_DIR <= '1';	-- Input
	F_DIR <= '0';	-- Output
	G_DIR <= '0';	-- Output
	G_LEV <= '0';	-- TTL

	HEART_BEAT_CNT_D <= HEART_BEAT_CNT_Q + 1;
	
	REFCLK_D <= not REFCLK_Q;

	C_DOUT(0) <= TRIGGER;
	F_DOUT(8) <= TRIGGER;

	SYNC <= SYNC_Q(0);
	SYNC_D <= "010" when SYNC_Q(0) = '1' else
	          "100" when SYNC_Q(1) = '1' else
	          "001";

	process(PLL_LOCK, PLLCLK)
	begin
		if PLL_LOCK = '0' then
			F_DOUT(7 downto 0) <= (others => '0');
			REFCLK_Q <= '0';
			G_DOUT(0) <= '0';
			G_DOUT(1) <= '0';
			SYNC_Q <= (others => '0');
		elsif rising_edge(PLLCLK) then
			SYNC_Q <= SYNC_D;
			REFCLK_Q <= REFCLK_D;
			F_DOUT(0) <= REFCLK_Q;
			F_DOUT(2) <= REFCLK_Q;
			F_DOUT(4) <= REFCLK_Q;
			F_DOUT(6) <= REFCLK_Q;
			F_DOUT(1) <= SYNC;
			F_DOUT(3) <= SYNC;
			F_DOUT(5) <= SYNC;
			F_DOUT(7) <= SYNC;
			
			G_DOUT(0) <= REFCLK_Q;		-- Test IO
			G_DOUT(1) <= SYNC;	-- Test IO
		end if;
	end process;
		
	------------ Unused IO ------------
	C_DOUT(31 downto 1) <= (others => '0');	-- Not used
	D_DOUT <= (others => '0');	-- Not used
	E_DOUT <= (others => '0');	-- Not used
	F_DOUT(31 downto 9) <= (others => '0');	-- Not used
--	F_DOUT(31 downto 8) <= (others => '0');	-- Not used
--	G_DOUT(1) <= '0';	-- Not used
	D_LEV <= '0';	-- Not used
	E_LEV <= '0';	-- Not used
	F_LEV <= '0';	-- Not used
	SPARE_OUT <= (others => '0');
	SPARE_DIR <= (others => '1');	
	PDL_WR <= '0';
	PDL_SEL <= '0';
	PDL_WRITE <= (others => '0');
	PDL_DIR <= '0';
	PDL0_IN <= '0';
	PDL1_IN <= '0';
	DLO0_GATE <= '0';
	DLO1_GATE <= '0';
	-----------------------------------

	PLL1: MPLLBlock
		port map(areset => SYS_RESET, inclk0 => LCLK, c0 => PLLCLK, locked => PLL_LOCK);
		
	TriggerDetect1: TriggerDetect
		port map(CLK => PLLCLK, RESET => RESET, SYNC => SYNC,
		         Q1_CLUSTERCNT1 => Q1_CLUSTERCNT1, Q1_CLUSTERCNT2 => Q1_CLUSTERCNT2, Q1_CLUSTERS => Q1_CLUSTERS,
		         Q2_CLUSTERCNT1 => Q2_CLUSTERCNT1, Q2_CLUSTERCNT2 => Q2_CLUSTERCNT2, Q2_CLUSTERS => Q2_CLUSTERS,
		         Q3_CLUSTERCNT1 => Q3_CLUSTERCNT1, Q3_CLUSTERCNT2 => Q3_CLUSTERCNT2, Q3_CLUSTERS => Q3_CLUSTERS,
		         Q4_CLUSTERCNT1 => Q4_CLUSTERCNT1, Q4_CLUSTERCNT2 => Q4_CLUSTERCNT2, Q4_CLUSTERS => Q4_CLUSTERS,
		         CLUSTERCNT_THR => CLUSTERCNT_THR, TRIGGER => TRIGGER);
		
	-- Quadrant 1 Instantiations
	ParallelComRx_Q1: ParallelComRx
		port map(CLK => PLLCLK, RESET => RESET, DIN => A_DIN, HITBITS => Q1_HITBITS, CLUSTERS => Q1_ICLUSTERS,
		         CLUSTERCNT1 => Q1_CLUSTERCNT1, CLUSTERCNT2 => Q1_CLUSTERCNT2, COMERROR => Q1_COMERROR, SYNC => SYNC);
	
	QISectSum_Q1: QISectSum
		port map(CLK => PLLCLK, QA => Q4_HITBITS, QB => Q1_HITBITS, QC => Q2_HITBITS, QSUMB => Q1_SUM);

	QISectClusters_Q1: QISectClusters
		port map(CLK => PLLCLK,	QSUMA => Q4_SUM, QSUMB => Q1_SUM, QSUMC => Q2_SUM,
		     THRESHOLD => CLUSTER_THRESHOLD, QICLUSTERSB => Q1_ICLUSTERS, QOCLUSTERSB => Q1_CLUSTERS);

	-- Quadrant 2 Instantiations
	ParallelComRx_Q2: ParallelComRx
		port map(CLK => PLLCLK, RESET => RESET, DIN => B_DIN, HITBITS => Q2_HITBITS, CLUSTERS => Q2_ICLUSTERS,
		         CLUSTERCNT1 => Q2_CLUSTERCNT1, CLUSTERCNT2 => Q2_CLUSTERCNT2, COMERROR => Q2_COMERROR, SYNC => SYNC);

	QISectSum_Q2: QISectSum
		port map(CLK => PLLCLK, QA => Q1_HITBITS, QB => Q2_HITBITS, QC => Q3_HITBITS, QSUMB => Q2_SUM);

	QISectClusters_Q2: QISectClusters
		port map(CLK => PLLCLK,	QSUMA => Q1_SUM, QSUMB => Q2_SUM, QSUMC => Q3_SUM,
		     THRESHOLD => CLUSTER_THRESHOLD, QICLUSTERSB => Q2_ICLUSTERS, QOCLUSTERSB => Q2_CLUSTERS);

	-- Quadrant 3 Instantiations
	ParallelComRx_Q3: ParallelComRx
		port map(CLK => PLLCLK, RESET => RESET, DIN => D_DIN, HITBITS => Q3_HITBITS, CLUSTERS => Q3_ICLUSTERS,
		         CLUSTERCNT1 => Q3_CLUSTERCNT1, CLUSTERCNT2 => Q3_CLUSTERCNT2, COMERROR => Q3_COMERROR, SYNC => SYNC);

	QISectSum_Q3: QISectSum
		port map(CLK => PLLCLK, QA => Q2_HITBITS, QB => Q3_HITBITS, QC => Q4_HITBITS, QSUMB => Q3_SUM);

	QISectClusters_Q3: QISectClusters
		port map(CLK => PLLCLK,	QSUMA => Q2_SUM, QSUMB => Q3_SUM, QSUMC => Q4_SUM,
		     THRESHOLD => CLUSTER_THRESHOLD, QICLUSTERSB => Q3_ICLUSTERS, QOCLUSTERSB => Q3_CLUSTERS);

	-- Quadrant 4 Instantiations
	ParallelComRx_Q4: ParallelComRx
		port map(CLK => PLLCLK, RESET => RESET, DIN => E_DIN, HITBITS => Q4_HITBITS, CLUSTERS => Q4_ICLUSTERS,
		         CLUSTERCNT1 => Q4_CLUSTERCNT1, CLUSTERCNT2 => Q4_CLUSTERCNT2, COMERROR => Q4_COMERROR, SYNC => SYNC);

	QISectSum_Q4: QISectSum
		port map(CLK => PLLCLK, QA => Q3_HITBITS, QB => Q4_HITBITS, QC => Q1_HITBITS, QSUMB => Q4_SUM);

	QISectClusters_Q4: QISectClusters
		port map(CLK => PLLCLK,	QSUMA => Q3_SUM, QSUMB => Q4_SUM, QSUMC => Q1_SUM,
		     THRESHOLD => CLUSTER_THRESHOLD, QICLUSTERSB => Q4_ICLUSTERS, QOCLUSTERSB => Q4_CLUSTERS);

	process(LCLK, SYS_RESET)
	begin
		if SYS_RESET = '1' then
			HEART_BEAT_CNT_Q <= (others => '0');
		elsif rising_edge(LCLK) then
			HEART_BEAT_CNT_Q <= HEART_BEAT_CNT_D;
		end if;
	end process;
	   
   --********************************************************************************
   -- USER REGISTERS SECTION
   --********************************************************************************
	
	-- WRITE REGISTERS
	P_WREG : process(LCLK, nLBRES)
	begin
		if (nLBRES = '0') then
			CLUSTER_THRESHOLD <= x"2";
			CLUSTERCNT_THR <= "0000010";
		elsif LCLK'event and LCLK = '1' then
			if (REG_WREN = '1') and (USR_ACCESS = '1') then
				case REG_ADDR is
					when A_THRESHOLD =>	CLUSTER_THRESHOLD <= REG_DIN(3 downto 0);
					when A_CLUSTCNT => CLUSTERCNT_THR <= REG_DIN(6 downto 0);
					when others =>		null;
				end case;
			end if;
		end if;
	end process;

	-- READ REGISTERS
	P_RREG: process(LCLK, nLBRES)
	begin
		if (nLBRES = '0') then
			REG_DOUT   <= (others => '0');
		elsif LCLK'event and LCLK = '1' then
			if (REG_RDEN = '1') and (USR_ACCESS = '1') then
				case REG_ADDR is
					when A_REVISION		=> REG_DOUT <= V_REVISION;
					when A_BOARDIDS		=> REG_DOUT <= V_MTRIGGERID & D_IDCODE & E_IDCODE & F_IDCODE;
					when A_ERRORS		=> REG_DOUT <= x"000" & CARD_D_ERROR & CARD_E_ERROR & CARD_F_ERROR & RESET;
					when others			=> REG_DOUT   <= (others => '0');
				end case;
			end if;
		end if;
	end process;    
   
END rtl;

