library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;  -- Use OR_REDUCE function

use work.v1495pkg.all;
use work.QTriggerPKG.all;

entity QTrigger is
	port( 
		nLBRES      : IN     std_logic;                       -- Async Reset (active low)
		LCLK        : IN     std_logic;                       -- Local Bus Clock
		--*************************************************
		-- REGISTER INTERFACE
		--*************************************************
		REG_WREN    : IN     std_logic;                       -- Write pulse (active high)
		REG_RDEN    : IN     std_logic;                       -- Read  pulse (active high)
		REG_ADDR    : IN     std_logic_vector (15 DOWNTO 0);  -- Register address
		REG_DIN     : IN     std_logic_vector (15 DOWNTO 0);  -- Data from CAEN Local Bus
		REG_DOUT    : OUT    std_logic_vector (15 DOWNTO 0);  -- Data to   CAEN Local Bus
		USR_ACCESS  : IN     std_logic;                       -- Current register access is 
													-- at user address space(Active high)
		--*************************************************
		-- V1495 Front Panel Ports (PORT A,B,C,G)
		--*************************************************
		A_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- In A (32 x LVDS/ECL)
		B_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- In B (32 x LVDS/ECL) 
		C_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- Out C (32 x LVDS)
		G_LEV       : OUT    std_logic;                       -- Output Level Select (NIM/TTL)
		G_DIR       : OUT    std_logic;                       -- Output Enable
		G_DOUT      : OUT    std_logic_vector (1 DOWNTO 0);   -- Out G - LEMO (2 x NIM/TTL)
		G_DIN       : IN     std_logic_vector (1 DOWNTO 0);   -- In G - LEMO (2 x NIM/TTL)
		--*************************************************
		-- A395x MEZZANINES INTERFACES (PORT D,E,F)
		--*************************************************
		-- Expansion Mezzanine Identifier:
		-- x_IDCODE :
		-- 000 : A395A (32 x IN LVDS/ECL)
		-- 001 : A395B (32 x OUT LVDS)
		-- 010 : A395C (32 x OUT ECL)
		-- 011 : A395D (8  x IN/OUT NIM/TTL)

		-- Expansion Mezzanine Port Signal Standard Select
		-- x_LEV : 
		--    0=>TTL,1=>NIM

		-- Expansion Mezzanine Port Direction
		-- x_DIR : 
		--    0=>OUT,1=>IN

		-- In/Out D (I/O Expansion)
		D_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- D slot mezzanine Identifier
		D_LEV       : OUT    std_logic;                       -- D slot Port Signal Level Select 
		D_DIR       : OUT    std_logic;                       -- D slot Port Direction
		D_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- D slot Data In  Bus
		D_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- D slot Data Out Bus
		-- In/Out E (I/O Expansion)
		E_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- E slot mezzanine Identifier
		E_LEV       : OUT    std_logic;                       -- E slot Port Signal Level Select
		E_DIR       : OUT    std_logic;                       -- E slot Port Direction
		E_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- E slot Data In  Bus
		E_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- E slot Data Out Bus
		-- In/Out F (I/O Expansion)
		F_IDCODE    : IN     std_logic_vector ( 2 DOWNTO 0);  -- F slot mezzanine Identifier
		F_LEV       : OUT    std_logic;                       -- F slot Port Signal Level Select
		F_DIR       : OUT    std_logic;                       -- F slot Port Direction
		F_DIN       : IN     std_logic_vector (31 DOWNTO 0);  -- F slot Data In  Bus
		F_DOUT      : OUT    std_logic_vector (31 DOWNTO 0);  -- F slot Data Out Bus
		--*************************************************
		-- DELAY LINES
		--*************************************************
		-- PDL = Programmable Delay Lines  (Step = 0.25ns / FSR = 64ns)
		-- DLO = Delay Line Oscillator     (Half Period ~ 10 ns)
		-- 3D3428 PDL (PROGRAMMABLE DELAY LINE) CONFIGURATION
		PDL_WR      : OUT    std_logic;                       -- Write Enable
		PDL_SEL     : OUT    std_logic;                       -- PDL Selection (0=>PDL0, 1=>PDL1)
		PDL_READ    : IN     std_logic_vector ( 7 DOWNTO 0);  -- Read Data
		PDL_WRITE   : OUT    std_logic_vector ( 7 DOWNTO 0);  -- Write Data
		PDL_DIR     : OUT    std_logic;                       -- Direction (0=>Write, 1=>Read)
		-- DELAY I/O
		PDL0_OUT    : IN     std_logic;                       -- Signal from PDL0 Output
		PDL1_OUT    : IN     std_logic;                       -- Signal from PDL1 Output
		DLO0_OUT    : IN     std_logic;                       -- Signal from DLO0 Output
		DLO1_OUT    : IN     std_logic;                       -- Signal from DLO1 Output
		PDL0_IN     : OUT    std_logic;                       -- Signal to   PDL0 Input
		PDL1_IN     : OUT    std_logic;                       -- Signal to   PDL1 Input
		DLO0_GATE   : OUT    std_logic;                       -- DLO0 Gate (active high)
		DLO1_GATE   : OUT    std_logic;                       -- DLO1 Gate (active high)
		--*************************************************
		-- SPARE PORTS
		--*************************************************
		SPARE_OUT    : OUT   std_logic_vector(11 downto 0);   -- SPARE Data Out 
		SPARE_IN     : IN    std_logic_vector(11 downto 0);   -- SPARE Data In
		SPARE_DIR    : OUT   std_logic_vector(11 downto 0);   -- SPARE Direction (0 => OUT, 1 => IN)   
		--*************************************************
		-- LED
		--*************************************************
		RED_PULSE       : OUT    std_logic;                   -- RED   Led Pulse (active high)
		GREEN_PULSE     : OUT    std_logic                    -- GREEN Led Pulse (active high)
	);
end QTrigger;

architecture rtl of qtrigger is

	component QPLLBlock
		port(
			areset		: IN STD_LOGIC  := '0';
			inclk0		: IN STD_LOGIC  := '0';
			c0			: OUT STD_LOGIC ;
			locked		: OUT STD_LOGIC 
		);
	end component;
	
	component ParallelCom
		port(
			CLK			: in std_logic;
			RESET		: in std_logic;
			SYNC		: in std_logic;
			BUSOUT		: out std_logic_vector(31 downto 0);
			QTOWERS		: in std_logic_vector(105 downto 0);
			QCLUSTERS	: in std_logic_vector(87 downto 0);
			CLUSTERCNT	: in std_logic_vector(7 downto 0);
			QSAMPLE		: out std_logic
		);
	end component;
	
	component QInputMask
		port(
			MASKS_IN	: in std_logic_vector(105 downto 0);
			QTOWERS_IN	: in std_logic_vector(105 downto 0);
			QTOWERS_OUT	: out std_logic_vector(105 downto 0);
			CONFIG0		: in std_logic_vector(105 downto 0);
			CONFIG1		: in std_logic_vector(105 downto 0)
		);
	end component;

	component FF
		port(
			aclr		: IN STD_LOGIC ;
			aset		: IN STD_LOGIC ;
			clock		: IN STD_LOGIC ;
			data		: IN STD_LOGIC ;
			q			: OUT STD_LOGIC 
		);
	end component;


	signal REVISION					: std_logic_vector(15 downto 0);

	signal PLL_REFCLK				: std_logic;
	signal SYNC_D, SYNC_Q			: std_logic;
	signal PLL_LOCK					: std_logic;
	signal SYS_RESET				: std_logic;	
	signal RESET					: std_logic;
	
	signal ERROR					: std_logic;
	signal CARD_D_ERROR				: std_logic;
	signal CARD_E_ERROR				: std_logic;
	signal CARD_F_ERROR				: std_logic;
	signal SYNC_ERROR				: std_logic;
	
	signal HEART_BEAT_CNT_D			: std_logic_vector(25 downto 0);
	signal HEART_BEAT_CNT_Q			: std_logic_vector(25 downto 0);
	
	signal PLLCLK					: std_logic;
	signal QTOWERS_D, QTOWERS_Q		: std_logic_vector(105 downto 0);
	
	signal QTOWERS_IN, QTOWERS_OUT	: std_logic_vector(105 downto 0);
	signal MASKS_IN					: std_logic_vector(105 downto 0);
	signal CONFIG0, CONFIG1			: std_logic_vector(105 downto 0);
	
	signal CLUSTER_THRESHOLD			: std_logic_vector(3 downto 0);
	signal WINDOWS						: STD_LOGIC_VEC_ARRAY9;
	signal WINDOW_SUM_D, WINDOW_SUM_Q	: STD_LOGIC_VEC_ARRAY4;
	signal QCLUSTERS_D, QCLUSTERS_Q		: std_logic_vector(87 downto 0);
	
	signal QSAMPLE, QSUM, QCLDETECT		: std_logic;
	signal CLUSTERCNT_D, CLUSTERCNT_Q	: std_logic_vector(7 downto 0);
	
	signal CLUSTERADDLIST1				: std_logic_vector(33 downto 0);
	signal CLUSTERADDLIST2				: std_logic_vector(68 downto 34);
	
	signal ADDR_MATCH_TOWERS			: std_logic_vector(6 downto 0);
	signal QTOWERS_HIT					: std_logic_vector(105 downto 0);
BEGIN

	SYS_RESET <= not nLBRES;
	RESET <= not PLL_LOCK;
	
	ERROR <= (not PLL_LOCK) or CARD_D_ERROR or CARD_E_ERROR or CARD_F_ERROR or SYNC_ERROR;

	CARD_D_ERROR <= '1' when D_IDCODE /= A395A else '0';
	CARD_E_ERROR <= '1' when E_IDCODE /= A395A else '0';
	CARD_F_ERROR <= '1' when E_IDCODE /= A395A else '0';
	
	SYNC_ERROR <= not (QSUM or QSAMPLE or QCLDETECT);

	RED_PULSE   <= ERROR;
	GREEN_PULSE <= HEART_BEAT_CNT_Q(25);

	D_LEV <= '0';	-- Not used
	E_LEV <= '0';	-- Not used
	F_LEV <= '0';	-- Not used
	G_LEV <= '0';	-- TTL

	D_DIR  <= '1';	-- Input
	E_DIR  <= '1';	-- Input
	F_DIR  <= '0';	-- Not used (output)
	G_DIR  <= '1';	-- Output

	D_DOUT <= (others => '0');	-- Not used
	E_DOUT <= (others => '0');	-- Not used
	F_DOUT <= (others => '0');	-- Not used
	G_DOUT <= (others => '0');	-- Not used

	HEART_BEAT_CNT_D <= HEART_BEAT_CNT_Q + 1;

	------------ Unused IO ------------
	SPARE_OUT <= (others => '0');
	SPARE_DIR <= (others => '1');	
	PDL_WR <= '0';
	PDL_SEL <= '0';
	PDL_WRITE <= (others => '0');
	PDL_DIR <= '0';
	PDL0_IN <= '0';
	PDL1_IN <= '0';
	DLO0_GATE <= '0';
	DLO1_GATE <= '0';
	-----------------------------------

	PLL_REFCLK <= G_DIN(0);
	
	SYNC_D <= G_DIN(1);
	
	MASKS_IN <= (others => '0');
	QTOWERS_IN <= E_DIN(9 downto 0) & D_DIN & B_DIN & A_DIN;
	QTOWERS_D <= QTOWERS_OUT when QSAMPLE = '1' else QTOWERS_Q;
	
	GEN1: for i in 0 to 105 generate
		DFF_GEN: FF	port map(
			aclr => ADDR_MATCH_TOWERS(i / 16),
			aset => QTOWERS_OUT(i),
			clock => '0',
			data => '0',
			q => QTOWERS_HIT(i)
	);
	end generate;
	
	PLL1: QPLLBlock
		port map(
			areset => SYS_RESET,
			inclk0 => PLL_REFCLK,
			c0 => PLLCLK,
			locked => PLL_LOCK
		);
	
	PCOM1: ParallelCom
		port map(
			CLK => PLLCLK,
			RESET => RESET,
			SYNC => SYNC_Q,
			BUSOUT => C_DOUT,
			QTOWERS => QTOWERS_Q,
			QCLUSTERS => QCLUSTERS_Q,
			CLUSTERCNT => CLUSTERCNT_Q,
			QSAMPLE => QSAMPLE
		);
		
	QINPUTMASK1: QInputMask
		port map(
			MASKS_IN => MASKS_IN,
			QTOWERS_IN => QTOWERS_IN,
			QTOWERS_OUT => QTOWERS_OUT,
			CONFIG0 => CONFIG0,
			CONFIG1 => CONFIG1
		);

	CLUSTERADDLIST1 <= QCLUSTERS_Q(40 downto 40) & QCLUSTERS_Q(38 downto 31) & QCLUSTERS_Q(29 downto 23) &
	                   QCLUSTERS_Q(21 downto 16) & QCLUSTERS_Q(14 downto 10) & QCLUSTERS_Q(8 downto 5) &
	                   QCLUSTERS_Q(3 downto 1);
	
	CLUSTERADDLIST2 <= QCLUSTERS_Q(78 downto 70) & QCLUSTERS_Q(68 downto 60) & QCLUSTERS_Q(58 downto 50) &
	                   QCLUSTERS_Q(48 downto 41);
	
	CLUSTERCNT_D(7 downto 0) <= stdvec_bitcnt(CLUSTERADDLIST2, 4) & stdvec_bitcnt(CLUSTERADDLIST1, 4);

	GEN2: for i in 0 to 87 generate
		WINDOWS(i) <= towerneighbors(i, QTOWERS_Q);
		WINDOW_SUM_D(i) <= sum_9bits(WINDOWS(i));
		QCLUSTERS_D(i) <= checkcluster(i, WINDOW_SUM_Q, CLUSTER_THRESHOLD);
	end generate;

	process(PLLCLK)
	begin
		if rising_edge(PLLCLK) then
			QSUM <= QSAMPLE;
			QCLDETECT <= QSUM;
		end if;
	end process;
	
	process(PLLCLK)
	begin
		if rising_edge(PLLCLK) then
			SYNC_Q <= SYNC_D;
			QTOWERS_Q <= QTOWERS_D;
			WINDOW_SUM_Q <= WINDOW_SUM_D;
			QCLUSTERS_Q <= QCLUSTERS_D;
			CLUSTERCNT_Q <= CLUSTERCNT_D;
		end if;
	end process;
	
	process(LCLK, SYS_RESET)
	begin
		if SYS_RESET = '1' then
			HEART_BEAT_CNT_Q <= (others => '0');
		elsif rising_edge(LCLK) then
			HEART_BEAT_CNT_Q <= HEART_BEAT_CNT_D;
		end if;
	end process;
	   
   --********************************************************************************
   -- USER REGISTERS SECTION
   --********************************************************************************

	-- WRITE REGISTERS
	P_WREG : process(LCLK, nLBRES)
	begin
		if (nLBRES = '0') then
			CLUSTER_THRESHOLD <= x"2";
			CONFIG0 <= (others => '0');
			CONFIG1 <= (others => '0');
		elsif LCLK'event and LCLK = '1' then
			if (REG_WREN = '1') and (USR_ACCESS = '1') then
				case REG_ADDR is
					when A_THRESHOLD	=> CLUSTER_THRESHOLD <= REG_DIN(3 downto 0);
					when A_CONFIG0_0	=> CONFIG0(15 downto 0) <= REG_DIN;
					when A_CONFIG0_1	=> CONFIG0(31 downto 16) <= REG_DIN;
					when A_CONFIG0_2	=> CONFIG0(47 downto 32) <= REG_DIN;
					when A_CONFIG0_3	=> CONFIG0(63 downto 48) <= REG_DIN;
					when A_CONFIG0_4	=> CONFIG0(79 downto 64) <= REG_DIN;
					when A_CONFIG0_5	=> CONFIG0(95 downto 80) <= REG_DIN;
					when A_CONFIG0_6	=> CONFIG0(105 downto 96) <= REG_DIN(9 downto 0);
					when A_CONFIG1_0	=> CONFIG1(15 downto 0) <= REG_DIN;
					when A_CONFIG1_1	=> CONFIG1(31 downto 16) <= REG_DIN;
					when A_CONFIG1_2	=> CONFIG1(47 downto 32) <= REG_DIN;
					when A_CONFIG1_3	=> CONFIG1(63 downto 48) <= REG_DIN;
					when A_CONFIG1_4	=> CONFIG1(79 downto 64) <= REG_DIN;
					when A_CONFIG1_5	=> CONFIG1(95 downto 80) <= REG_DIN;
					when A_CONFIG1_6	=> CONFIG1(105 downto 96) <= REG_DIN(9 downto 0);
					when others			=> null;
				end case;
			end if;
		end if;
	end process;

	-- READ REGISTERS
	P_RREG: process(LCLK, nLBRES)
	begin
		if (nLBRES = '0') then
			REG_DOUT   <= (others => '0');
		elsif LCLK'event and LCLK = '1' then
			ADDR_MATCH_TOWERS <= (others => '0');
			if (REG_RDEN = '1') and (USR_ACCESS = '1') then
				case REG_ADDR is
					when A_REVISION		=> REG_DOUT <= V_REVISION;
					when A_BOARDIDS		=> REG_DOUT <= V_QTRIGGERID & D_IDCODE & E_IDCODE & F_IDCODE;
					when A_ERRORS		=> REG_DOUT <= x"000" & CARD_D_ERROR & CARD_E_ERROR & CARD_F_ERROR & RESET;
					when A_TOWERS_0		=> REG_DOUT <= QTOWERS_HIT(15 downto 0); ADDR_MATCH_TOWERS(0) <= '1';
					when A_TOWERS_1		=> REG_DOUT <= QTOWERS_HIT(31 downto 16); ADDR_MATCH_TOWERS(1) <= '1';
					when A_TOWERS_2		=> REG_DOUT <= QTOWERS_HIT(47 downto 32); ADDR_MATCH_TOWERS(2) <= '1';
					when A_TOWERS_3		=> REG_DOUT <= QTOWERS_HIT(63 downto 48); ADDR_MATCH_TOWERS(3) <= '1';
					when A_TOWERS_4		=> REG_DOUT <= QTOWERS_HIT(79 downto 64); ADDR_MATCH_TOWERS(4) <= '1';
					when A_TOWERS_5		=> REG_DOUT <= QTOWERS_HIT(95 downto 80); ADDR_MATCH_TOWERS(5) <= '1';
					when A_TOWERS_6		=> REG_DOUT <= "000000" & QTOWERS_HIT(105 downto 96); ADDR_MATCH_TOWERS(6) <= '1';
					when others			=> REG_DOUT <= (others => '0');
				end case;
			end if;
		end if;
	end process;    
   
END rtl;

